@@ -24,7 +24,10 @@
 
 #include <sys/time.h>
 #include <time.h>
+
+#ifndef __MINGW__
 #include <pwd.h>
+#endif
 
 #ifdef USE_TERMCAP
 #include <curses.h>
@@ -51,7 +54,7 @@
 
 /* ---------------------------------------------------------------------- */
 
-/** doesn't currently appear to be used anywhere? */
+/** [TODO get rid?] doesn't currently appear to be used anywhere? */
 #define USE_OLD_COMPOSE  
 
 /* ---------------------------------------------------------------------- */
@@ -59,6 +62,8 @@
 /**
  * Convert string function for ASCII mode.
  *
+ * This is used for the "printToken" function in the relevant PDR.
+ *
  * @param s  The string to convert.
  * @return   s (ie no change).
  */
@@ -69,16 +74,14 @@ ascii_convert_string(char *s)
   return s;
 }
 
-char *
-ascii_print_field(FieldType field, int at_end);
+char *ascii_print_field(FieldType field, int at_end);
 
 /* ---------------------------------------------------------------------- */
 
 /**
  * Print description record for ASCII print mode.
  */
-PrintDescriptionRecord 
-ASCIIPrintDescriptionRecord = {
+PrintDescriptionRecord ASCIIPrintDescriptionRecord = {
   "%9d: ",                            /* CPOSPrintFormat */
   
   NULL,                               /* BeforePrintStructures */
@@ -112,8 +115,7 @@ ASCIIPrintDescriptionRecord = {
 /**
  * Print description record for Highlighted-ASCII print mode.
  */
-PrintDescriptionRecord 
-ASCIIHighlightedPrintDescriptionRecord = {
+PrintDescriptionRecord ASCIIHighlightedPrintDescriptionRecord = {
   "%9d: ",                            /* CPOSPrintFormat */
   
   NULL,                               /* BeforePrintStructures */
@@ -183,7 +185,8 @@ get_colour_escape(char colour, int foreground)
  * Dummy function
  */
 char *
-get_typeface_escape(char typeface) {
+get_typeface_escape(char typeface)
+{
   return "";
 }
 
@@ -201,8 +204,8 @@ void get_screen_escapes(void)
   sc_bl_in = NULL;
   sc_bl_out = NULL;
   sc_all_out = NULL;
- 
-  return;
+
+  escapes_initialized++;
 }
 
 #else /* USE_TERMCAP */
@@ -317,7 +320,7 @@ get_typeface_escape(char typeface)
   case 's': return sc_s_in;
   case 'n': return sc_all_out;        /* also switches off colour */
   default:
-    fprintf(stderr, "Internal error: unknown typeface '%c'.\n", typeface);
+    Rprintf("Internal error: unknown typeface '%c'.\n", typeface);
     return "";
   }
 }
@@ -339,7 +342,7 @@ get_colour_escape(char colour, int foreground) {
       case 'p': return "\x1B[0;35m";
       case 'c': return "\x1B[0;36m";
       default:
-        fprintf(stderr, "Internal error: unknown colour '%c'.\n", colour);
+        Rprintf("Internal error: unknown colour '%c'.\n", colour);
         return "\x1B[0m";
       }
     }
@@ -352,7 +355,7 @@ get_colour_escape(char colour, int foreground) {
       case 'p': return "\x1B[0;45m";
       case 'c': return "\x1B[0;46m";
       default:
-        fprintf(stderr, "Internal error: unknown colour '%c'.\n", colour);
+        Rprintf("Internal error: unknown colour '%c'.\n", colour);
         return "\x1B[0m";
       }
     }
@@ -381,7 +384,7 @@ char *
 ascii_print_field(FieldType field, int at_end)
 {
 
-  *sc_before_token = 0;                /* sets sc_before_token to "" */
+  sc_before_token[0] = '\0';                /* sets sc_before_token to "" */
 
   /* if targets are shown, print target number at end of target/keyword fields */
   if (show_targets && at_end && (field==TargetField || field==KeywordField)) {
@@ -445,6 +448,16 @@ ascii_print_field(FieldType field, int at_end)
   return sc_before_token;
 }
 
+/**
+ * Prints a line of text (which will have been previously exrtracted from a corfpus
+ * linked to the present corpus by an a-attribute) with a brief character-mode
+ * start-of-line flag ("-->$att_name: ").
+ *
+ * @param stream          Destination for the output.
+ * @param highlighting    Boolean: if true, use colour/bold highlighting for the leading indicator on the line.
+ * @param attribute_name  The name of the aligned corpus: printed in the leading indicator
+ * @param line            Character data of the line of aligned-corpus data to print. This is treated as opaque.
+ */
 void
 ascii_print_aligned_line(FILE *stream, 
                          int highlighting,
@@ -455,21 +468,22 @@ ascii_print_aligned_line(FILE *stream,
     char *red = get_colour_escape('r', 1);
     char *bold = get_typeface_escape('b');
     char *normal = get_typeface_escape('n');
-    fprintf(stream, "%s%s-->%s:%s %s\n", 
+    Rprintf("%s%s-->%s:%s %s\n", 
             red, bold,
             attribute_name,
             normal,
             line);
   }
   else
-    fprintf(stream, "-->%s: %s\n", attribute_name, line);
+    Rprintf("-->%s: %s\n", attribute_name, line);
 }
 
 
 /* print the concordance line for the target_word on the screen */
 /**
  * Prints a concordance line.
- * (documentation not compklete)_
+ * (documentation not complete)_
+ *
  *
  */
 void 
@@ -569,10 +583,14 @@ ascii_print_corpus_header(CorpusList *cl,
                           FILE *stream)
 {
   time_t now;
+
+#ifndef __MINGW__
   struct passwd *pwd = NULL;
+#endif
+
   int i;
   
-  (void) time(&now);
+  time(&now);
   /*   pwd = getpwuid(geteuid()); */
   /* disabled because of incompatibilities between different Linux versions */
 
@@ -581,21 +599,26 @@ ascii_print_corpus_header(CorpusList *cl,
     fputc('-', stream);
   fputc('\n', stream);
   
-  fprintf(stream,
+  Rprintf(
           "#\n"
           "# User:    %s (%s)\n"
           "# Date:    %s"
           "# Corpus:  %s (%s)\n"
           "# Name:    %s:%s\n"
           "# Size:    %d intervals/matches\n",
+#ifndef __MINGW__
           (pwd ? pwd->pw_name : "<unknown>"),
           (pwd ? pwd->pw_gecos  : "<unknown>"),
+#else
+          "<unknown>",
+          "<unknown>",
+#endif
           ctime(&now),
           (cl->corpus && cl->corpus->registry_name ? cl->corpus->registry_name : "<Unknown Corpus>"),
           (cl->corpus && cl->corpus->name ? cl->corpus->name : "<Unknown Corpus>"),
           cl->mother_name, cl->name,
           cl->size);
-  fprintf(stream,
+  Rprintf(
           "# Context: %d %s left, %d %s right\n"
           "#\n",
           CD.left_width,
@@ -608,7 +631,7 @@ ascii_print_corpus_header(CorpusList *cl,
            (CD.right_structure_name) ? CD.right_structure_name : "???"));
   
   if (cl->query_corpus && cl->query_text) {
-    fprintf(stream, "# Query: %s; %s\n", cl->query_corpus, cl->query_text);
+    Rprintf("# Query: %s; %s\n", cl->query_corpus, cl->query_text);
   }
   
   
@@ -633,7 +656,7 @@ ascii_print_output(CorpusList *cl,
   if ((last >= cl->size) || (last < 0))
     last = cl->size - 1;
 
-  for (i = first; (i <= last) && (!broken_pipe); i++) {
+  for (i = first; (i <= last) && !cl_broken_pipe; i++) {
     
     if (cl->sortidx)
       real_line = cl->sortidx[i];
@@ -641,7 +664,7 @@ ascii_print_output(CorpusList *cl,
       real_line = i;
 
     if (GlobalPrintOptions.number_lines) {
-      fprintf(outfd, "%6d.\t", output_line);
+      Rprintf("%6d.\t", output_line);
       output_line++;
     }
 
@@ -667,7 +690,7 @@ ascii_print_group(Group *group, int expand, FILE *fd)
   last_source_id = -666;
   nr_targets = 0;
 
-  for (cell = 0; cell < group->nr_cells; cell++) {
+  for (cell = 0; (cell < group->nr_cells) && !cl_broken_pipe; cell++) {
 
     source_id = group->count_cells[cell].s;
     source_s = Group_id2str(group, source_id, 0);
@@ -682,20 +705,20 @@ ascii_print_group(Group *group, int expand, FILE *fd)
         nr_targets = 0;
       }
 
-      /* separator bar is meaningless when using the internal grouping algorithm */
-      if ((UseExternalGrouping) || (cell == 0))
-        fprintf(fd, SEPARATOR);
+      /* separator bar between groups */
+      if (cell == 0 || (group->is_grouped && nr_targets == 0))
+        Rprintf(SEPARATOR);
       
-      fprintf(fd, "%-28s  %-28s\t%6d\n",
+      Rprintf("%-28s  %-28s\t%6d\n",
               (nr_targets == 0) ? source_s : " ", target_s, count);
     }
     else {
       if (source_id < 0) source_s = "";        /* don't print "(none)" or "(all)" in plain mode (just empty string) */
       if (target_id < 0) target_s = "";
       if (has_source) 
-        fprintf(fd, "%s\t%s\t%d\n", source_s, target_s, count);
+        Rprintf("%s\t%s\t%d\n", source_s, target_s, count);
       else 
-        fprintf(fd, "%s\t%d\n", target_s, count);
+        Rprintf("%s\t%d\n", target_s, count);
     }
     
     if (expand) {
