@@ -15,8 +15,6 @@
  *  WWW at http://www.gnu.org/copyleft/gpl.html).
  */
 
-void Rprintf(const char *, ...);
-
 #include <stddef.h>
 #include <stdio.h>
 #include <sys/types.h>
@@ -39,15 +37,14 @@ void Rprintf(const char *, ...);
 #include "options.h"
 #include "output.h"
 #include "ranges.h"
-#include "paths.h"
 
 #define COLON ':'
 #define SLASH '^'
 
 
-#define subcorpload_debug False
+#define subcorpload_debug 0
 
-/** magic number for {?? subcorpus files} */
+/** magic number for subcorpus (incl. query) file format */
 #define SUBCORPMAGIC 36193928
 /* the sum of the original programmers' birthdays: 15081963 (Max) + 21111965 (Oli) */
 
@@ -114,8 +111,7 @@ initialize_cl(CorpusList *cl, int free_name)
   cl->corpus = NULL;
   cl->size = 0;
 
-  if (cl->cd) 
-    FreeContextDescriptor(&(cl->cd));
+  cl_free(cl->cd);
 
   cl_free(cl->sortidx);
 
@@ -183,6 +179,10 @@ NewCL(void)
 
 /* ------------------------------------------------------------ */
 
+/**
+ * Returns a FieldType enumeration corresponding to the field name
+ * indicated by its stirng argument.
+ */
 FieldType 
 field_name_to_type(char *name)
 {
@@ -201,6 +201,10 @@ field_name_to_type(char *name)
     return NoField;
 }
 
+/**
+ * Returns a pointer to an internal constant string that labels
+ * the FieldType argument.
+ */
 char *
 field_type_to_name(FieldType ft) {
   switch (ft) {
@@ -220,7 +224,19 @@ field_type_to_name(FieldType ft) {
   }
 }
 
-
+/**
+ * Counts the number of field-value items of a specified type in
+ * the given subcorpus (that is, query resultset).
+ *
+ * If the type is MatchField, then the N of values is simply equal
+ * to the number of query results. If it is KeywordField or
+ * TargetField, the number returned is the number of results
+ * where the field exists (which is not always all of them).
+ *
+ * @param cl  The query result to analyse.
+ * @param ft  The field type to count.
+ * @return    The number of values of the speciifed field-type.
+ */
 int 
 NrFieldValues(CorpusList *cl, FieldType ft)
 {
@@ -250,7 +266,7 @@ NrFieldValues(CorpusList *cl, FieldType ft)
       
     default:
     case NoField:
-      fprintf(stderr, "Illegal field type %d\n", ft);
+      Rprintf("Illegal field type %d\n", ft);
       break;
     }
   }
@@ -259,7 +275,7 @@ NrFieldValues(CorpusList *cl, FieldType ft)
 
 /* ---------------------------------------------------------------------- */
 
-/* A utility function required by ensure_corpus_size() */
+/** A utility function required by ensure_corpus_size() */
 int
 SystemCorpusSize(Corpus *corpus)
 {
@@ -340,7 +356,7 @@ ensure_corpus_size(CorpusList *cl)
     if (!cl->loaded) {
       /* load subcorpus (the local_dir entry of the corpus structure contains
          the name of the directory where the disk file can be found */
-      char filename[1024];
+      char filename[CL_MAX_FILENAME_LENGTH];
       
       /* re-create subcorpus filename from corpus structure
          (cf. the treatment in load_corpusnames()) */
@@ -436,19 +452,39 @@ LoadedCorpus(char *name,
 
 /* ---------------------------------------------------------------------- */
 
+/**
+ * Finds the pointer to the corpus (or subcorpus, or query result) with the given name.
+ *
+ * When searching for s (name of corpus) strcmp() is used; no
+ * case conversion is done.
+ *
+ * If "type" is UNDEF, it returns the first
+ * corpus with matching name. Otherwise the returned corpus
+ * has the type "type".
+ *
+ * @param s                     name of the corpus to find (as string)
+ * @param type                  If this is UNDEF, all corpora are checked;
+ *                              if it is any other type, only corproa of that
+ *                              type are checked.
+ * @param try_recursive_search  Boolean: whether or not to try to find corpus
+ *                              through implicit expansion.
+ * @return                      Pointer to the CorpusList object for the
+ *                              specified corpus. NULL is returned when the
+ *                              corpus is not yet present.
+ */
 CorpusList *
 findcorpus(char *s, CorpusType type, int try_recursive_search)
 {
   CorpusList *sp, *tmp;
   char *colon;
 
-  char mother_corpus[64];
+  char mother_corpus[CL_MAX_LINE_LENGTH];
   char *expansion;
   char *real_name;
 
   if (type != SYSTEM) {
     if ((colon = strchr(s, COLON)) != NULL) {
-      strcpy(mother_corpus, s);
+      cl_strcpy(mother_corpus, s);
       mother_corpus[colon - s] = '\0';
       real_name = colon + 1;
     }
@@ -464,10 +500,15 @@ findcorpus(char *s, CorpusType type, int try_recursive_search)
 
   expansion = strchr(real_name, SLASH);
 
-  if (mother_corpus[0] == '\0')
-    sp = LoadedCorpus(real_name, NULL, type);
-  else
-    sp = LoadedCorpus(real_name, mother_corpus, type);
+  if (expansion) {
+    sp = NULL; /* don't re-use existing subcorpus generated by implicit expansion */
+  }
+  else {
+    if (mother_corpus[0] == '\0')
+      sp = LoadedCorpus(real_name, NULL, type);
+    else
+      sp = LoadedCorpus(real_name, mother_corpus, type);
+  }
 
   /* We need to call ensure_corpus_size() to implement delayed loading. */
   if (sp) {
@@ -477,10 +518,10 @@ findcorpus(char *s, CorpusType type, int try_recursive_search)
   /* (sp==0): try to find corpus through implicit expansion ("A^s" == "A expand to s") */
   else if (type != SYSTEM && expansion && try_recursive_search) {
       
-    char new_name[128];
+    char new_name[CL_MAX_LINE_LENGTH];
     Context ctx;
 
-    strcpy(new_name, real_name);
+    cl_strcpy(new_name, real_name);
     new_name[expansion - real_name] = '\0';
 
     if (mother_corpus[0] == '\0')
@@ -496,12 +537,15 @@ findcorpus(char *s, CorpusType type, int try_recursive_search)
         return NULL; 
       if (!access_corpus(sp))
         return NULL;
+      if (sp->type == SYSTEM) {
+        cqpmessage(Warning, "Implicit expansion %s only allowed for named query result.", s);
+        return NULL;
+      }
       
       assert(sp->corpus);
       
       if ((ctx.attrib = find_attribute(sp->corpus, expansion+1, ATT_STRUC, NULL)) == NULL) {
-        cqpmessage(Warning, "Can't expand to ``%s'' -- \n"
-                   "no such structural attribute in %s",
+        cqpmessage(Warning, "Can't expand to <%s> regions -- no such structural attribute in %s",
                    expansion+1, sp->mother_name);
         return NULL;
       }
@@ -509,11 +553,10 @@ findcorpus(char *s, CorpusType type, int try_recursive_search)
       ctx.type = structure;
       ctx.size = 1;
 
-
-      tmp = duplicate_corpus(sp, real_name, False);
+      tmp = duplicate_corpus(sp, real_name, True);
       
       if (tmp == NULL) {
-        fprintf(stderr, "Weird error in findcorpus\n");
+        Rprintf("Internal error in findcorpus() -- this should not happen!\n");
         return NULL;
       }
 
@@ -547,6 +590,12 @@ findcorpus(char *s, CorpusType type, int try_recursive_search)
   return sp;
 }
 
+/**
+ * Remove a corpus from the global list of corpora.
+ *
+ * @see       corpuslist
+ * @param cl  The corpus to drop.
+ */
 void
 dropcorpus(CorpusList *cl)
 {
@@ -563,8 +612,7 @@ dropcorpus(CorpusList *cl)
       prev = prev->next;
 
     if (prev == NULL) {
-      fprintf(stderr, "%s:dropcorpus(): cl is not in list of loaded corpora\n",
-              __FILE__);
+      Rprintf("%s:dropcorpus(): cl is not in list of loaded corpora\n", __FILE__);
       cl = NULL;
     }
     else {
@@ -573,8 +621,7 @@ dropcorpus(CorpusList *cl)
     }
   }
   else {
-    fprintf(stderr, "%s:dropcorpus(): cl is not in list of loaded corpora (list empty)\n",
-            __FILE__);
+    Rprintf("%s:dropcorpus(): cl is not in list of loaded corpora (list empty)\n", __FILE__);
     cl = NULL;
   }
 
@@ -587,6 +634,25 @@ dropcorpus(CorpusList *cl)
   }
 }
 
+/**
+ * Duplicate a corpus via its CorpusList object.
+ *
+ * duplicate_corpus creates a copy of an existing corpus
+ * and casts its type to SUB. The new corpus is given the
+ * name "new_name". If a subcorpus of that name is already
+ * present, NULL is retured if force_overwrite is False. If
+ * force_overwrite is True, the old corpus is discarded.
+ *
+ * @param cl               The corpus to duplicate
+ * @param new_name         Name for the duplicated corpus.
+ * @param force_overwrite  Boolean: whether or not to force an
+ *                         overwrite if the subcorpus you are
+ *                         attempting to create already exists.
+ * @return                 NULL if you attempted to overwrite
+ *                         with force_overwrite == false.
+ *                         Otherwise, a CorpusList pointer to
+ *                         the new corpus.
+ */
 CorpusList *
 duplicate_corpus(CorpusList *cl,
                  char *new_name,
@@ -595,8 +661,7 @@ duplicate_corpus(CorpusList *cl,
   CorpusList *newc;
 
   if (cl == NULL) {
-    fprintf(stderr, "%s:duplicate_corpus(): WARNING: Called with NULL corpus\n",
-            __FILE__);
+    Rprintf("%s:duplicate_corpus(): WARNING: Called with NULL corpus\n", __FILE__);
     return NULL;
   }
 
@@ -704,6 +769,19 @@ duplicate_corpus(CorpusList *cl,
   return newc;
 }
 
+/**
+ * Copy a corpus as type TEMP.
+ *
+ * make_temp_corpus makes a copy of the corpus in *cl
+ * into a corpus of type "TEMP" with name "new_name".
+ * If a temporary corpus with that name already exists,
+ * it is overwritten.
+ *
+ * @param cl         The corpus to copy.
+ * @param new_name   Name for the temporary copy.
+ * @return           NULL for error. Otherwise, a
+ *                   CorpusList pointer to the new corpus.
+ */
 CorpusList *
 make_temp_corpus(CorpusList *cl,
                  char *new_name)
@@ -711,7 +789,7 @@ make_temp_corpus(CorpusList *cl,
   CorpusList *newc;
 
   if (cl == NULL) {
-    fprintf(stderr, "%s:duplicate_corpus(): WARNING: Called with NULL corpus\n",
+    Rprintf("%s:duplicate_corpus(): WARNING: Called with NULL corpus\n",
             __FILE__);
     return NULL;
   }
@@ -797,13 +875,26 @@ make_temp_corpus(CorpusList *cl,
   return newc;
 }
 
+/**
+ * Convert a temporary corpus to a real subcorpus.
+ *
+ * assign_temp_to_sub assigns the temporary corpus in *tmp
+ * to a "real" subcorpus with name "subname". If such a
+ * subcorpus already exists, it is overwritten. The temporary
+ * corpus is deleted afterwards. The return value is the new
+ * subcorpus (which may be equal to tmp, but not necessarily).
+ *
+ * @param tmp      Temporary corpus to convert.
+ * @param subname  Name to use for new subcorpus.
+ * @return         Pointer to new subcorpus.
+ */
 CorpusList *
 assign_temp_to_sub(CorpusList *tmp, char *subname)
 {
   CorpusList *cl;
 
   if (tmp == NULL) {
-    fprintf(stderr, "%s:duplicate_corpus(): WARNING: Called with NULL corpus\n",
+    Rprintf("%s:duplicate_corpus(): WARNING: Called with NULL corpus\n",
             __FILE__);
     return NULL;
   }
@@ -870,6 +961,12 @@ assign_temp_to_sub(CorpusList *tmp, char *subname)
   return NULL;
 }
 
+/**
+ * Delete temproary corpora.
+ *
+ * drop_temp_corpora clears the list of corpora of all
+ * temporary stuff.
+ */
 void
 drop_temp_corpora(void)
 {
@@ -880,6 +977,7 @@ drop_temp_corpora(void)
    * since dropcorpus does the very same linear search
    * too), but keep this until I have some spare
    * time
+   * TODO
    */
   
   prev = NULL;
@@ -913,27 +1011,46 @@ drop_temp_corpora(void)
 
 /* ---------------------------------------------------------------------- */
 
+/**
+ * Creates a copy of the string with the given mode (LOWER/UPPER)
+ * enforced on it.
+ */
 static char *
-changecase_string(char *str, enum case_mode mode)
+changecase_string(const char *str, enum case_mode mode)
 {
   char *str_new;
-  int i, len = strlen(str);
+  /*int i, len = strlen(str);*/
 
   str_new = cl_strdup(str);
 
-  for (i = 0; i <= len; i++)
+  /*
+   for (i = 0; i <= len; i++)
     str_new[i] = (mode == LOWER) ? tolower(str[i]) : toupper(str[i]);
+  */
+  if (mode == LOWER)
+    cl_id_tolower(str_new);
+  else
+    cl_id_toupper(str_new);
 
-  return (str_new);
+  return str_new;
 }
 
+/**
+ * Like changecase_string(), but modifies the string in situ.
+ */
 static char *
 changecase_string_no_copy(char *str, enum case_mode mode)
 {
-  int i;
+  /*int i;*/
 
+  /*
   for (i = 0; str[i]; i++)
     str[i] = (mode == LOWER) ? tolower(str[i]) : toupper(str[i]);
+  */
+  if (mode == LOWER)
+    cl_id_tolower(str);
+  else
+    cl_id_toupper(str);
 
   return str;
 }
@@ -941,7 +1058,7 @@ changecase_string_no_copy(char *str, enum case_mode mode)
 static char *
 get_fulllocalpath(CorpusList *cl, int qualify)
 {
-  char fullname[1024];
+  char fullname[CL_MAX_FILENAME_LENGTH];
   char *upname;
 
   if (qualify) {
@@ -965,6 +1082,7 @@ get_fulllocalpath(CorpusList *cl, int qualify)
 /* ---------------------------------------------------------------------- */
 
 
+/* TODO change function name. "accessible" is ri-bloody-diculous */
 /**
  * Tests whether a file is accessible.
  *
@@ -1015,12 +1133,12 @@ int
 check_stamp(char *directory, char *fname)
 {
   FILE *fd;
-  char full_name[1024];
+  char full_name[CL_MAX_FILENAME_LENGTH];
   int magic, ok;
 
-  sprintf(full_name, "%s/%s", directory, fname);
+  sprintf(full_name, "%s" SUBDIR_SEP_STRING "%s", directory, fname);
 
-  if (((fd = OpenFile(full_name, "r")) == NULL) ||
+  if (((fd = open_file(full_name, "rb")) == NULL) ||
       (fread(&magic, sizeof(int), 1, fd) == 0) ||
       ((magic != SUBCORPMAGIC) && (magic != SUBCORPMAGIC+1)))
     ok = 0;
@@ -1042,19 +1160,18 @@ load_corpusnames(enum corpus_type ct)
   struct dirent *ep;
   char          *entry;
 
-  char          dirlist[MAX_LINE_LENGTH];
+  char          dirlist[CL_MAX_LINE_LENGTH];
   char          *dirname;
   CorpusList    *corpus;
 
   if (!((ct == SYSTEM) || (ct == SUB))) {
-    fprintf(stderr, "Can't load corpus names for type %d\n",
-            ct);
+    Rprintf("Can't load corpus names for type %d\n", ct);
     return;
   }
 
   if (ct == SYSTEM) {
     if (registry == NULL)
-      strcpy(dirlist, central_corpus_directory());
+      strcpy(dirlist, cl_standard_registry());
     else
       strcpy(dirlist, registry);
   }
@@ -1073,7 +1190,7 @@ load_corpusnames(enum corpus_type ct)
     if (*dirname == '?') {
       dirname++;
       optional_dir = 1;
-  }
+    }
 
     dp = opendir(dirname);
   
@@ -1107,7 +1224,7 @@ load_corpusnames(enum corpus_type ct)
               /*     if (corpus == NULL) { */
               /* (NB: one data directory constraint is implicit; loading might work, but save_subcorpus() will crash miserably) */
 
-              char mother[1024]; /* Judith vs. Oli, round 231 */
+              char mother[CL_MAX_FILENAME_LENGTH]; /* Judith vs. Oli, round 231 */
               
               /* allocate memory for the new id */
               corpus = NewCL();
@@ -1155,7 +1272,7 @@ load_corpusnames(enum corpus_type ct)
     }
     else if (!silent && !optional_dir)
       cqpmessage(Warning, "Couldn't open directory %s (continuing)", dirname);
-  }
+  } /* end for (each directory) */
 }
 
 void
@@ -1208,7 +1325,7 @@ GetSystemCorpus(char *name, char *registry)
     else if (registry)
       cl->registry = cl_strdup(registry);
     else {
-      fprintf(stderr, "Warning: no registry directory for %s\n",
+      Rprintf("Warning: no registry directory for %s\n",
               name);
       cl->registry = NULL;
     }
@@ -1275,7 +1392,12 @@ ensure_syscorpus(char *registry, char *name)
 }
 
 
-
+/**
+ * @param cl
+ * @param advertised_directory
+ * @param advertised_filename
+ * @return                      Boolean: whether the file was loaded correctly.
+ */
 static Boolean
 attach_subcorpus(CorpusList *cl,
                  char *advertised_directory,
@@ -1299,7 +1421,7 @@ attach_subcorpus(CorpusList *cl,
 
     if (advertised_directory && advertised_filename) {
 
-      char sname[1024];
+      char sname[CL_MAX_FILENAME_LENGTH];
       
       strcpy(sname, advertised_directory);
       if (sname[strlen(sname)-1] != '/')
@@ -1311,20 +1433,20 @@ attach_subcorpus(CorpusList *cl,
     else
       fullname = get_fulllocalpath(cl, 0);
 
-    if ((fp = OpenFile(fullname, "r")) == NULL && !advertised_filename) {
+    if ((fp = open_file(fullname, "rb")) == NULL && !advertised_filename) {
       cl_free(fullname);
       fullname = get_fulllocalpath(cl, 1);
-      fp = OpenFile(fullname, "r");
+      fp = open_file(fullname, "rb");
     }
 
     if (fp == NULL)
-      fprintf(stderr, "Subcorpus %s not accessible (can't open %s for reading)\n",
+      Rprintf("Subcorpus %s not accessible (can't open %s for reading)\n",
               cl->name, fullname);
     else {
       len = file_length(fullname);
 
       if (len <= 0)
-        fprintf(stderr, "ERROR: File length of subcorpus is <= 0\n");
+        Rprintf("ERROR: File length of subcorpus is <= 0\n");
       else {
 
 
@@ -1334,9 +1456,9 @@ attach_subcorpus(CorpusList *cl,
         /* read the subcorpus */
         
         if (len != fread(field, 1, len, fp))
-          fprintf(stderr, "Read error while reading subcorpus %s\n", cl->name);
+          Rprintf("Read error while reading subcorpus %s\n", cl->name);
         else if ((*((int *)field) != SUBCORPMAGIC) && (*((int *)field) != SUBCORPMAGIC+1))
-          fprintf(stderr, "Magic number incorrect in %s\n", fullname);
+          Rprintf("Magic number incorrect in %s\n", fullname);
         else {
 
           CorpusList *mother;
@@ -1351,6 +1473,7 @@ attach_subcorpus(CorpusList *cl,
           cl->abs_fn = fullname;
           fullname = NULL;
           
+          /* advance p to the end of the 1st string (path of mother's registry) */
           while (*p)
             p++;
           /* skip the '\0' character */
@@ -1371,7 +1494,7 @@ attach_subcorpus(CorpusList *cl,
 
             assert(cl->mother_size > 0);
 
-            /* advance p to the end of the 2nd string */
+            /* advance p to the end of the 2nd string (name of mother) */
             while (*p)
               p++;
             /* skip the '\0' character */
@@ -1442,7 +1565,7 @@ attach_subcorpus(CorpusList *cl,
             }
 
             if (subcorpload_debug) {
-              fprintf(stderr,
+              Rprintf(
                       "Header size: %ld\n"
                       "Nr Matches: %d\n"
                       "regdir: %s\n"
@@ -1452,7 +1575,7 @@ attach_subcorpus(CorpusList *cl,
                       cl->registry,
                       cl->mother_name);
               for (j = 0; j < cl->size; j++)
-                fprintf(stderr, 
+                Rprintf(
                         "range[%d].start = %d\n"
                         "range[%d].end   = %d\n",
                         j, cl->range[j].start, j, cl->range[j].end);
@@ -1488,7 +1611,7 @@ save_subcorpus(CorpusList *cl, char *fname)
   int i, l1, l2, magic;
 
   FILE *fp;
-  char outfn[1024];
+  char outfn[CL_MAX_FILENAME_LENGTH];
 
   if (cl == NULL)
     return False;
@@ -1512,8 +1635,9 @@ save_subcorpus(CorpusList *cl, char *fname)
           return False;
         }
 
-        sprintf(outfn, "%s/%s:%s", 
+        sprintf(outfn, "%s%c%s:%s",
                 LOCAL_CORP_PATH, 
+                SUBDIR_SEPARATOR,
                 cl->mother_name ? cl->mother_name : "NONE",
                 cl->name);
 
@@ -1521,7 +1645,7 @@ save_subcorpus(CorpusList *cl, char *fname)
       }
     }
 
-    if ((fp = OpenFile(fname, "w")) != NULL) {
+    if ((fp = open_file(fname, "wb")) != NULL) {
 
       int zero; 
       zero = 0;
@@ -1587,7 +1711,7 @@ save_subcorpus(CorpusList *cl, char *fname)
       return(True);
     }
     else {
-      fprintf(stderr, "cannot open output file %s\n", fname);
+      Rprintf("cannot open output file %s\n", fname);
       return(False);
     }
   }
@@ -1639,6 +1763,15 @@ NextCorpusFromList(CorpusList *cl)
 }
 
 
+/**
+ * Assesses whether a specified corpus can be accessed.
+ *
+ * That is, it makes sure that the data for corpus in
+ * "cl" is loaded and accessible.
+ *
+ * @param cl  A CorpusList specifying the corpus to check.
+ * @return    A boolean - true if cl can be accessed.
+ */
 Boolean
 access_corpus(CorpusList *cl)
 {
@@ -1668,8 +1801,15 @@ access_corpus(CorpusList *cl)
 }
 
 
-/* implements the search strategy for corpora */
-
+/**
+ * Find the CorpusList object corresponding to a corpus name.
+ *
+ * First the SUB corpora (created by queries) are searched,
+ * then the SYSTEM corproa.
+ *
+ * @param name  String containing name of corpus to find.
+ * @return      Pointer to desired CorpusList.
+ */
 CorpusList *
 search_corpus(char *name)
 {
@@ -1682,10 +1822,21 @@ search_corpus(char *name)
   return cl;
 }
 
-/* change_corpus takes as argument a corpus name and makes 
- * the corpus accessible
+/**
+ * Make a corpus accessible for searching as the "current" corpus.
+ *
+ * change_corpus sets the current corpus to the corpus with
+ * name "name", first searching SUB corpora, then searching
+ * SYSTEM corpora.
+ *
+ * When a corpus is "made accessible", its name is checked for validity and
+ * availability; if all is OK, set_current_corpus is called on it.
+ *
+ * @param name    A string indicating the name of a corpus.
+ * @param silent  Boolean. Ignored.
+ * @return        Boolean. True if the corpus was set successfully, otherwise
+ *                false.
  */
-
 Boolean
 change_corpus(char *name, Boolean silent)
 {
@@ -1693,7 +1844,7 @@ change_corpus(char *name, Boolean silent)
 
   if ((cl = search_corpus(name)) != NULL) {
     if (!access_corpus(cl)) {
-      fprintf(stderr, "Can't access corpus %s, keep previous corpus\n",
+      Rprintf("Can't access corpus %s, keep previous corpus\n",
               cl->name);
       cl = NULL;
     }
@@ -1708,18 +1859,22 @@ change_corpus(char *name, Boolean silent)
 }
 
 Boolean
-valid_subcorpus_id(char *corpusname) {
+valid_subcorpus_id(char *corpusname)
+{
   return (findcorpus(corpusname, SYSTEM, 0) ? False : True);
 }
 
+/** Checks whether corpusname is syntactically valid as a query result name */
 Boolean
-valid_subcorpus_name(char *corpusname) {
-  char mother[MAX_LINE_LENGTH];
-  return ((split_subcorpus_name(corpusname, mother) == NULL) ? False : True);
+valid_subcorpus_name(char *corpusname)
+{
+  return ((split_subcorpus_name(corpusname, NULL) == NULL) ? False : True);
 }
 
+/** Checks whether corpusname is fully qualified (with name of mother corpus); does not imply syntatic validity */
 Boolean
-is_qualified(char *corpusname) {
+is_qualified(char *corpusname)
+{
   return (strchr(corpusname, COLON) == NULL ? 0 : 1);
 }
 
@@ -1729,7 +1884,7 @@ is_qualified(char *corpusname) {
  *
  * This function splits query result name {corpusname} into qualifier (name of mother corpus) and local name;
  * returns pointer to local name part, or NULL if {corpusname} is not syntactically valid;
- * if mother_name is not NULL, it must point to a buffer of suitable length (MAX_LINE_LENGTH is sufficient)
+ * if mother_name is not NULL, it must point to a buffer of suitable length (CL_MAX_LINE_LENGTH is sufficient)
  * where the qualifier will be stored (empty string for unqualified corpus, and return value == {corpusname} in this case)
  */
 char *
@@ -1778,15 +1933,14 @@ split_subcorpus_name(char *corpusname, char *mother_name)
  * @param cp  The corpus to touch. This must be of type SUB.
  * @return    Boolean: true if the touch worked, otherwise false.
  */
-Boolean
+int
 touch_corpus(CorpusList *cp)
 {
-  if ((!cp) ||
-      (cp->type != SUB))
-    return False;
+  if ((!cp) || (cp->type != SUB))
+    return 0;
   else {
-    cp->saved = False;
-    cp->needs_update = True;
+    cp->saved = 0;
+    cp->needs_update = 1;
 
     return 1;
   }
@@ -1796,7 +1950,7 @@ touch_corpus(CorpusList *cp)
 /**
  * Sets the current corpus (by pointer to the corpus).
  *
- * Also, execustes Xkwic side effects, if necessary
+ * Also, executes Xkwic side effects, if necessary.
  *
  * @param cp     Pointer to the corpus to set as current.
  *               cp may be NULL, which is legal.
@@ -1860,6 +2014,8 @@ set_current_corpus_name(char *name, int force)
 
 /**
  * Internal function for sorting list of corpus names.
+ *
+ * @see show_corpora_files
  */
 static int
 show_corpora_files_sort(const void *p1, const void *p2)
@@ -1870,28 +2026,38 @@ show_corpora_files_sort(const void *p1, const void *p2)
   return result;
 }
 
+/**
+ * Function that does the work for show_corpora_files
+ *
+ * @see show_corpora_files
+ */
 void 
 show_corpora_files1(enum corpus_type ct)
 {
   CorpusList *cl;
-  char **list;                  /* list of corpus names */
+  char **list;                  /* list of corpus names (for qsort) */
   int i, N;
-  char initial = ' ';           /* inital character of corpus name (insert <br> before new initial letter) */
+  char initial = ' ';           /* inital character of corpus name
+                                 * (for use in case of indented list;
+                                 * insert <br> before new initial letter) */
   char label[4];
 
   if (ct == SYSTEM) {
-    if (pretty_print) Rprintf("System corpora:\n");
     /* make list of corpus names, then qsort() and print */
-    N = 0;                      /* count nr of system corpora before allocating list */
-    for (cl = corpuslist; cl; cl = cl->next) 
-      if (cl->type == SYSTEM)  N++;
+    /* count nr of system corpora before allocating list */
+    for (cl = corpuslist, N = 0; cl; cl = cl->next)
+      if (cl->type == SYSTEM)
+        N++;
     list = (char **) cl_malloc(N * sizeof(char *)); /* allocate list */
     for (cl = corpuslist, i = 0; cl; cl = cl->next) /* compile list of corpus names */
-      if (cl->type == SYSTEM)  list[i++] = cl->name;
+      if (cl->type == SYSTEM)
+        list[i++] = cl->name;
     qsort(list, N, sizeof(char *), show_corpora_files_sort);
 
-    if (pretty_print) 
+    if (pretty_print) {
+      Rprintf("System corpora:\n");
       start_indented_list(0,0,0);       /* now print sorted list */
+    }
     for (i = 0; i < N; i++) {
       if (pretty_print) {
         if (list[i][0] != initial) {
@@ -1911,32 +2077,36 @@ show_corpora_files1(enum corpus_type ct)
     free(list);
   }
   else if (ct == SUB) {
-    if (pretty_print) Rprintf("Named Query Results:\n");
+    if (pretty_print)
+      Rprintf("Named Query Results:\n");
     for (cl = corpuslist; cl; cl = cl->next)
-      if (cl->type == SUB) {
-        if (pretty_print) {
-          Rprintf("   %c%c%c  %s:%s [%d]\n", 
+      if (cl->type == SUB)
+          Rprintf(pretty_print ? "   %c%c%c  %s:%s [%d]\n" : "%c%c%c\t%s:%s\t%d\n",
                  cl->loaded ? 'm' : '-',
                  cl->saved ? 'd' : '-',
                  cl->needs_update ? '*' : '-',
                  cl->mother_name ? cl->mother_name : "???",
-                 cl->name, cl->size);
-        }
-        else {
-          Rprintf("%c%c%c\t%s:%s\t%d\n", 
-                 cl->loaded ? 'm' : '-',
-                 cl->saved ? 'd' : '-',
-                 cl->needs_update ? '*' : '-',
-                 cl->mother_name ? cl->mother_name : "???",
-                 cl->name, cl->size);
-        }
-      }
+                 cl->name,
+                 cl->size);
   }
   else {
     assert(0 && "Invalid argument in show_corpora_files()<corpmanag.c>.");
   }
 }
 
+/**
+ * A function to print out a list of corpora currently available.
+ *
+ * "files" is a misnomer; it actually looks on the global list of currently
+ * loaded corpora, and prints their names.
+ *
+ * Either system corpora (SYSTEM) or subcorpora (SUB) can be shown, depending on ct. If
+ * ct is UNDEF, both are shown.
+ *
+ * For subcorpora, a bundle of other information is shown too.
+ *
+ * @param ct  Type of corpus to show (SUB, SYSTEM or UNDEF).
+ */
 void 
 show_corpora_files(enum corpus_type ct)
 {
@@ -1948,120 +2118,3 @@ show_corpora_files(enum corpus_type ct)
     show_corpora_files1(ct);
 }
 
-
-
-/* NEW FUNCTIONS ================================================== */
-
-/*
- * TODO  (??)
- *
- * None of these functions is used anywhere.
- * Their names are not noticeably better than the original names!
- * Also, they are a deeply inefficient way of rewriting the API; all those unnecessary function calls!
- * Can they be got rid of? -- and maybe rewrite the API the way SE did it in the CL instead -- AH. 26.9.09
- */
-
-
-/* ============================== THE LIST OF CORPORA */
-
-void
-CorpusListInit(void)
-{
-  init_corpuslist();
-}
-
-void CorpusListFree(void)
-{
-  free_corpuslist();
-}
-
-/* ============================== CORPUS NAMES */
-
-Boolean
-CorpusNameValid(char *name) {
-  return valid_subcorpus_id(name);
-}
-
-Boolean
-CorpusNameQualified(char *name) {
-  return is_qualified(name);
-}
-
-/* ============================== OPERATIONS ON/WITH CORPORA */
-
-void
-CorpusLoadDescriptors(CorpusType ct)
-{
-  check_available_corpora(ct);
-}
-
-Boolean CorpusDiscard(CorpusList *cl,
-                      Boolean remove_file_also, 
-                      Boolean save_if_unsaved)
-{
-  dropcorpus(cl);
-
-  return True;
-}
-
-Boolean CorpusSave(CorpusList *cl, char *file_name)
-{
-  return save_subcorpus(cl, file_name);
-}
-
-Boolean CorpusSaveAll(void)
-{
-  save_unsaved_subcorpora();
-  return True;
-}
-
-Boolean CorpusDiscardTMPCorpora(void)
-{
-  drop_temp_corpora();
-
-  return True;
-}
-
-CorpusList *CorpusDuplicate(CorpusList *cl,
-                            char *new_name,
-                            Boolean force_overwrite)
-{
-  return duplicate_corpus(cl, new_name, force_overwrite);
-}
-
-CorpusList *CorpusDuplicateIntoTMP(CorpusList *cl,
-                                   char *new_name)
-{
-  return make_temp_corpus(cl, new_name);
-}
-
-CorpusList *CorpusChangeTMPtoSUB(CorpusList *tmp, char *subname)
-{
-  return assign_temp_to_sub(tmp, subname);
-}
-
-Boolean CorpusLoad(CorpusList *cl)
-{
-  return access_corpus(cl);
-}
-
-Boolean CorpusSetCurrentByname(char *name)
-{
-  return set_current_corpus_name(name, 0);
-}
-
-Boolean CorpusSetCurrent(CorpusList *cl)
-{
-  return set_current_corpus(cl, 0);
-}
-
-Boolean CorpusTouch(CorpusList *cl)
-{
-  return touch_corpus(cl);
-}
-
-void CorpusShowNames(CorpusType ct)
-{
-  show_corpora_files(ct);
-}
-
