@@ -26,8 +26,8 @@
 
 /** Level of progress-info (inc compression protocol) message output: 0 = none. */
 int do_protocol = 0;
-/** File handle for this program's progress-info output: always stdout */
-FILE *protocol; /* " = stdout;" init moved to main() for Gnuwin32 compatibility */
+/** File handle for this program's progress-info output: note, it is always stdout */
+FILE *protocol; /* For Gnuwin32 compatibility, this must be initialised in main(), not here. */
 
 /* ---------------------------------------------------------------------- */
 
@@ -38,12 +38,13 @@ char *corpus_id = NULL;
 
 int debug = 0;
 
-void usage(char *msg, int error_code);
+void huffcode_usage(char *msg, int error_code);
 
 /* ---------------------------------------------------------------------- */
 
 /**
- * Prints a binary representation of an integer to a stream.
+ * Prints, to the specified stream, a string containing
+ * a binary representation of an integer.
  *
  * @param i       Integer to print
  * @param width   Number of bits in the integer
@@ -94,7 +95,7 @@ bprintf(unsigned int i, int width, FILE *stream)
 
 
 /**
- * Dumps the specified heap of memory to the program output stream.
+ * Dumps the specified heap of memory to the program's STDOUT.
  *
  * @see protocol
  * @param heap       Location of the heap to dump.
@@ -124,7 +125,7 @@ dump_heap(int *heap, int heap_size, int node, int indent)
 }
 
 /**
- * Prints a description of the specified heap of memory to the program output stream.
+ * Prints a description of the specified heap of memory to the program's STDOUT.
  *
  * @see protocol
  * @param heap       Location of the heap to print.
@@ -144,7 +145,7 @@ print_heap(int *heap, int heap_size, char *title)
   
   dump_heap(heap, heap_size, 1, 0);
 
-  fprintf(protocol, "");
+  fprintf(protocol, "\x0c");
 }
 
 
@@ -170,21 +171,18 @@ sift(int *heap, int heap_size, int node)
    * 2i+1. So we maintain this scheme and decrement just before
    * addressing the array.
    *
-   * left child in 2*node, right child in 2*node + 1, parent in
-   * node */
+   * left child in 2*node, right child in 2*node + 1, parent in node
+   */
 
   while (child <= heap_size) {
  
     if ((child < heap_size) && 
         (heap[heap[child]] < heap[heap[child-1]])) {
-      
       /* select right branch in heap[child+1-1] */
-      
       child++;
     }
     
     if (heap[heap[node-1]] > heap[heap[child-1]]) {
-      
       /* root is larger than selected child, so we have to swap and
        * recurse down */
 
@@ -195,7 +193,6 @@ sift(int *heap, int heap_size, int node)
       heap[node-1] = tmp;
       
       /* recurse downwards */
-      
       node = child;
       child = node * 2;
     }
@@ -219,7 +216,7 @@ WriteHCD(char *filename, HCD *hc)
 {
   FILE *fd;
 
-  if ((fd = fopen(filename, "w")) == NULL) {
+  if ((fd = fopen(filename, "wb")) == NULL) {
     perror(filename);
     return 0;
   }
@@ -242,10 +239,10 @@ WriteHCD(char *filename, HCD *hc)
 }
 
 /**
- * Reads a Huffman compressed sequence from file.
+ * Reads a Huffman Code Descriptor from file.
  *
- * @param filename  Path to file where compressed sequence is saved.
- * @param hc        Pointer to location where the sequence's descriptor block will be loaded to.
+ * @param filename  Path to file where descriptor is saved.
+ * @param hc        Pointer to location where the descriptor block will be loaded to.
  * @return          Boolean: true for all OK, false for error.
  */
 int
@@ -253,7 +250,7 @@ ReadHCD(char *filename, HCD *hc)
 {
   FILE *fd;
 
-  if ((fd = fopen(filename, "r")) == NULL) {
+  if ((fd = fopen(filename, "rb")) == NULL) {
     perror(filename);
     return 0;
   }
@@ -273,7 +270,8 @@ ReadHCD(char *filename, HCD *hc)
     return 1;
   }
 }
-/* should these two functions perhaps be in cl/attributes.h? (prototype of HCD is in attributes.h) or all HCD object in separate module? */
+/* TODO: should these two functions perhaps be in cl/attributes.h? (prototype of HCD is in attributes.h)
+ *       or all HCD object in separate module? */
 
 
 /* ================================================== COMPRESSION */
@@ -295,8 +293,9 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
   int nr_codes = 0;
 
-  int *heap = NULL;
-  unsigned *codelength = NULL;        /* was char[], probably to save space; but that's unnecessary and makes gcc complain */
+  unsigned int *heap = NULL;
+  /* must be unsigned: because of add-1 trick below to avoid codes longer than 32 bits, cumulative frequencies in heap may exceed 2G limit */
+  unsigned int *codelength = NULL;
 
   int issued_codes[MAXCODELEN];
   int next_code[MAXCODELEN];
@@ -338,12 +337,10 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
               corpus->registry_dir, corpus->registry_name, attr->any.name);
       exit(1);
     }
-
   }
 
   /*
-   * strongly follows Witten/Moffat/Bell: ``Managing Gigabytes'', 
-   * pp. 335ff.
+   * strongly follows Witten/Moffat/Bell: ``Managing Gigabytes'', pp. 335ff.
    */
 
   hc->size = cl_max_id(attr);                /* the size of the attribute (nr of items) */
@@ -362,28 +359,39 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
   hc->min_codelen = 100;
   hc->max_codelen = 0;
 
-  bzero((char *)hc->lcount, MAXCODELEN * sizeof(int));
-  bzero((char *)hc->min_code, MAXCODELEN * sizeof(int));
-  bzero((char *)hc->symindex, MAXCODELEN * sizeof(int));
+  memset((char *)hc->lcount,   '\0', MAXCODELEN * sizeof(int));
+  memset((char *)hc->min_code, '\0', MAXCODELEN * sizeof(int));
+  memset((char *)hc->symindex, '\0', MAXCODELEN * sizeof(int));
 
-  bzero((char *)issued_codes, MAXCODELEN * sizeof(int));
+  memset((char *)issued_codes, '\0', MAXCODELEN * sizeof(int));
 
-  codelength = (unsigned *)cl_calloc(hc->size, sizeof(unsigned));
+  codelength = (unsigned int *)cl_calloc(hc->size, sizeof(unsigned));
 
 
   /* =========================================== make & initialize the heap */
 
-  heap = (int *)cl_malloc(hc->size * 2 * sizeof(int));
+  heap = (unsigned int *)cl_malloc(hc->size * 2 * sizeof(int));
+  /* The heap initially consists of two consecutive arrays:
+   * a) an array of pointers into the second array (given as offsets on heap[])
+   * b) an array of frequency counts for the leaves of the Huffman tree
+   * In the algorithm below, inner nodes of the tree are created by merging the two least frequent
+   * nodes. The corresponding frequency counts, which are no longer needed, are replaced by pointers
+   * to the parent node.
+   */
 
   for (i = 0; i < hc->size; i++) {
     heap[i] = hc->size + i;
-    heap[hc->size+i] = get_id_frequency(attr, i);
+    heap[hc->size+i] = cl_id2freq(attr, i) + 1;
+    /* add-one trick needed to avoid unsupported Huffman codes > 31 bits for very large corpora of ca. 2 billion words:
+     * theoretical optimal code length for hapax legomena in such corpora is ca. 31 bits, and the Huffman algorithm
+     * sometimes generates 32-bit codes; with add-one trick, the theoretical optimal code length is always <= 30 bits
+     * NB: this means that cumulative frequency counts may exceed the corpus size and hence the 2G limit (-> unsigned int)
+     */
   }
 
   /* ============================== PROTOCOL ============================== */
   if (do_protocol > 0)
-    fprintf(protocol, "Allocated heap with %d cells for %d items\n\n",
-            hc->size * 2, hc->size);
+    fprintf(protocol, "Allocated heap with %d cells for %d items\n\n", hc->size * 2, hc->size);
   if (do_protocol > 2)
     print_heap(heap, hc->size, "After Initialization");
   /* ============================== PROTOCOL ============================== */
@@ -399,7 +407,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
    * we address the heap in the following manner: when we start array
    * indices at 1, the left child is at 2i, and the right child is at
    * 2i+1. So we maintain this scheme and decrement just before
-   * adressing the array. 
+   * addressing the array.
    */
 
   /*
@@ -412,7 +420,6 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
      * bottom up, left to right,
      * for each root of each subtree, sift if necessary
      */
-
     sift(heap, h, i);
   }
 
@@ -509,8 +516,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
   /* ================================================== Phase 3 */
 
-  /* compute the code lengths. We don't have any freqs in heap any
-   * more, only pointers to parents */
+  /* compute the code lengths. We don't have any freqs in heap any more, only pointers to parents */
 
   heap[0] = -1U;
 
@@ -531,16 +537,16 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
   for (i = 0; i < hc->size; i++) {
 
     int cl = heap[i+hc->size];
+    if (cl == 0) {
+      assert((hc->size == 1) && "Major error: code length of 0 bits should only happen for lexicon size = 1");
+      cl = 1; /* special case: if lexicon contains only a single type, generate 1-bit code '0' */
+    }
 
     sum_bits += cl * get_id_frequency(attr, i);
-
     codelength[i] = cl;
-    if (cl == 0)
-      continue;
 
     if (cl > hc->max_codelen)
       hc->max_codelen = cl;
-
     if (cl < hc->min_codelen)
       hc->min_codelen = cl;
 
@@ -568,7 +574,6 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
     fprintf(stderr, "Problem: No output generated -- no items?\n");
     nr_codes = 0;
-
   }
   else {
 
@@ -607,7 +612,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
     /* ============================== PROTOCOL ============================== */
     if (do_protocol > 1) {
-      fprintf(protocol, "\n");
+      fprintf(protocol, "\n");
       fprintf(protocol, "   Item   f(item)  CL      Bits     Code, String\n");
       fprintf(protocol, "------------------------------------"
               "------------------------------------\n");
@@ -657,9 +662,9 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
     {
       char *path;
 
-      char hcd_path[MAX_LINE_LENGTH];
-      char huf_path[MAX_LINE_LENGTH];
-      char sync_path[MAX_LINE_LENGTH];
+      char hcd_path[CL_MAX_LINE_LENGTH];
+      char huf_path[CL_MAX_LINE_LENGTH];
+      char sync_path[CL_MAX_LINE_LENGTH];
 
       Component *corp;
 
@@ -695,8 +700,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
       printf("- writing code descriptor block to %s\n",  hcd_path);
       if (!WriteHCD(hcd_path, hc)) {
-        fprintf(stderr, "ERROR: writing %s failed. Aborted.\n",
-                hcd_path);
+        fprintf(stderr, "ERROR: writing %s failed. Aborted.\n", hcd_path);
         exit(1);
       }
 
@@ -719,7 +723,6 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
       for (i = 0; i < hc->length; i++) {
 
         /* SYNCHRONIZE */
-
         if ((i % SYNCHRONIZATION) == 0) {
           if (i > 0)
             BFflush(&bfd);
@@ -732,9 +735,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
           cdperror("(aborting) cl_cpos2id() failed");
           exit(1);
         }
-
         else {
-
           assert((id >= 0) && (id < hc->size) && "Internal Error");
 
           cl = codelength[id];
@@ -745,9 +746,7 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
                     id, code, cl, i);
             exit(1);
           }
-
         }
-
       }
 
       fclose(sync);
@@ -755,8 +754,8 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
     }
   }
 
-  free(codelength);
-  free(heap);
+  cl_free(codelength);
+  cl_free(heap);
  
   return 1;
 }
@@ -765,9 +764,6 @@ compute_code_lengths(Attribute *attr, HCD *hc, char *fname)
 
 /* ================================================== DECOMPRESSION & ERROR CHECKING */
 
-/* this
-    */
-
 /**
  * Checks a huffcoded attribute for errors by decompressing it.
  *
@@ -793,9 +789,9 @@ decode_check_huff(Attribute *attr, char *fname)
   
   unsigned char bit;
 
-  char hcd_path[MAX_LINE_LENGTH];
-  char huf_path[MAX_LINE_LENGTH];
-  char sync_path[MAX_LINE_LENGTH];
+  char hcd_path[CL_MAX_LINE_LENGTH];
+  char huf_path[CL_MAX_LINE_LENGTH];
+  char sync_path[CL_MAX_LINE_LENGTH];
 
   
   printf("VALIDATING %s.%s\n", corpus_id, attr->any.name);
@@ -898,7 +894,8 @@ decode_check_huff(Attribute *attr, char *fname)
   printf("!! You can delete the file <%s> now.\n",
          component_full_name(attr, CompCorpus, NULL));
   
-  return;                        /* exits on error, so there's no return value */
+  return;
+  /* exits on error, so there's no return value */
 }
 
 
@@ -911,7 +908,7 @@ decode_check_huff(Attribute *attr, char *fname)
  * @param error_code  Value to be returned by the program when it exits.
  */
 void 
-usage(char *msg, int error_code)
+huffcode_usage(char *msg, int error_code)
 {
   if (msg)
     fprintf(stderr, "Usage error: %s\n", msg);
@@ -927,13 +924,13 @@ usage(char *msg, int error_code)
   fprintf(stderr, "  -r <dir>  set registry directory\n");
   fprintf(stderr, "  -f <file> set output file prefix (creates <file>.huf, ...)\n");
   fprintf(stderr, "  -v        verbose mode (shows protocol) [may be repeated]\n");
-/*   fprintf(stderr, "  -d        debug mode (not implemented)\n"); */
+/*   fprintf(stderr, "  -d        debug mode (not implemented)\n"); *//* TODO -d / -D distinct as in cwb-compress-rdx? */
   fprintf(stderr, "  -T        skip validation pass ('I trust you')\n");
   fprintf(stderr, "  -h        this help page\n\n");
   fprintf(stderr, "Part of the IMS Open Corpus Workbench v" VERSION "\n\n");
 
   if (corpus)
-    drop_corpus(corpus);
+    cl_delete_corpus(corpus);
 
   exit(error_code);
 }
@@ -949,7 +946,8 @@ usage(char *msg, int error_code)
  * @param argv   Command-line arguments.
  */
 int 
-main(int argc, char **argv) {
+main(int argc, char **argv)
+{
   char *registry_directory = NULL;
   char *output_fn = NULL;
   char *attr_name = DEFAULT_ATT_NAME;
@@ -971,7 +969,7 @@ main(int argc, char **argv) {
   progname = argv[0];
 
   /* parse arguments */
-  while ((c = getopt(argc, argv, "+TvP:r:f:dAh")) != EOF)
+  while ((c = getopt(argc, argv, "+TvP:r:f:dAh")) != EOF) {
     switch (c) {
 
       /* T: skip decompression / error checking pass ("I trust you")  */
@@ -1016,24 +1014,25 @@ main(int argc, char **argv) {
 
       /* h: help page */
     case 'h':
-      usage(NULL, 2);
+      huffcode_usage(NULL, 2);
       break;
 
     default: 
-      usage("illegal option.", 2);
+      huffcode_usage("illegal option.", 2);
       break;
     }
+  }
   
   /* single argument: corpus id */
   if (optind < argc) {
     corpus_id = argv[optind++];
   }
   else {
-    usage("corpus not specified (missing argument)", 1);
+    huffcode_usage("corpus not specified (missing argument)", 1);
   }
 
   if (optind < argc) {
-    usage("Too many arguments", 1);
+    huffcode_usage("Too many arguments", 1);
   }
   
   if ((corpus = cl_new_corpus(registry_directory, corpus_id)) == NULL) {
