@@ -1,11 +1,11 @@
 library(RcppCWB)
 
-testthat::context("cwb_cpos2str")
+testthat::context("cl_cpos2str")
 
 test_that(
   "cpos2str",
   {
-    token <- cwb_cpos2str(
+    token <- cl_cpos2str(
       corpus = "REUTERS",
       p_attribute = "word",
       registry = system.file(package = "RcppCWB", "extdata", "cwb", "registry"),
