@@ -1,11 +1,11 @@
 library(RcppCWB)
 
-testthat::context("cwb_id2freq")
+testthat::context("cl_id2freq")
 
 test_that(
   "id2freq",
   {
-    N <- cwb_id2freq(
+    N <- cl_id2freq(
       corpus = "REUTERS",
       p_attribute = "word",
       id = 15,
